#pragma once

/*
Copyright (c) 2013-2014 Gustav Hartvigsson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include "baseobject.h"
#include "defs.h"
#include "utils.h"
#include "Func.h"
#include "DynamicArray.h"
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

S_BEGIN_DECLS

/**
 * @file
 * @defgroup SError SError
 * @addtogroup SError
 * @{
 *
 * Error that can be added to and passed down to the caller.
 *
 * @section Theory
 *
 * The Theory behind how SError works is a complex story in where there are
 * dragons.
 *
 */

#define S_ERROR_DOMAIN_RANGE_MAX 128

typedef schar * (* SErrorDomainToString)(sint error_id, schar * error_msg);

#define S_ERROR_DOMAIN_TO_STRING_FUNC(k) ((SErrorDomainToString)(k))

/**
 * An SError is a data structure that inherits from SBaseObjectInstance.
 *
 * An SError represents an error that can occur.
 */
typedef struct SError SError;

typedef sint SErrorDomain;


#define S_ERROR(o) (SError *)(o)
#define S_ERROR_CLASS(k) (SErrorClass *)(k)

/** @brief
 * The different error types.
 *
 * These are the erros that are in the default error domain.
 */
typedef enum {
  S_ERROR_NONE = 0, /**< Not on error */
  S_ERROR_INPUT_OUTPUT,/**< An I/O error */
  S_ERROR_OVERFLOW, /**< An Overflow error */
  S_ERROR_TYPE_ERROR, /** Some generic type error. */
  S_ERROR_UNUSED_0,
  S_ERROR_UNUSED_1,
  S_ERROR_UNUSED_2,
  S_ERROR_OTHER, /**< Some unknowned error */
  S_ERROR_NULL /**< An NULL error */
} SDefaultErrorType;

/**
 * @var static char * SDefaultErrorTypeName[]
 * @breif
 * Holds the names of the string version of SDefaultErrorType.
 */
static schar * SDefaultErrorTypeName[] S_UNUSED = {
  "NONE",
  "INPUT/OUTPUT",
  "OVERFLOW",
  "TYPE_ERROR",
  "UNUSED_0/INVALID",
  "UNUSED_1/INVALID",
  "UNUSED_2/INVALID",
  "OTHER",
  "NULL",
  0x0,
  0x0
};

S_EXPORTED
schar *
s_default_error_domain_to_string (sint error_id, schar * msg);

/**
 * The Default Error Domain.
 */
#define S_ERROR_GET_DEFAULT_DOMAIN\
  s_error_get_domain ("default",\
                      S_ERROR_DOMAIN_TO_STRING_FUNC (s_default_error_domain_to_string))

/**
 * @brief
 * Function to get the same of an error.
 *
 * For use in bindings.
 *
 * @param k The key to look up.
 */
S_EXPORTED
schar *
s_default_error_get_name (SDefaultErrorType k);

/**
 * the constructor for the an SError.
 *
 */
S_EXPORTED
SError *
s_error_new (void);

/**
 * This function calles the deinitize method of the object.
 */
S_EXPORTED
void
s_error_free (SError * self);

S_EXPORTED
void
s_error_append (SError * self ,sint error, const schar * message, SErrorDomain error_domain);

S_EXPORTED
sboolean
s_error_has_error (SError * self);

/**
 * Returns an array with the error strings.
 * Must be freed by caller.
 */
S_EXPORTED
SDynamicArray *
s_error_to_string_array (SError * self);


/**
 * This function should be run at the end of each program.
 * It frees any resorses that has been allocated in the running of the program.
 */
S_EXPORTED
void
s_error_teardown (void);

/**
 * This function prints the current error to stdout.
 */
S_EXPORTED
void
s_error_print_error (SError * self);


/**
 * This function registers an error domain if it has not allready been registered,
 * and if it has been registered it returns the SErrorDomain handle associated
 * with the error domain.
 *
 * @param name
 *
 * @param to_string_func the
 *
 * @returns an SErrorDomain handle.
 *
 * @returns < 0 an error
 * has ocured. Most likely a missmatch between the provided to_string_func
 * and a to_string_func that allready has been registered.
 *
 * To make this lees tedious to work with it is recomended to make a macro
 * to call this function.
 @code {.c}
#define MY_MODULE_ERROR_DOMAIN\
  s_error_get_domain ("MyModule"\
                      my_module_error_to_string_func)
 @endcode
 */
S_EXPORTED
SErrorDomain
s_error_get_domain (const schar * name,
                    SErrorDomainToString to_string_func);



/** @} */

S_END_DECLS
